import abc


class AbstractExampleParser(abc.ABC):
    """Abstract example parser"""

    @abc.abstractmethod
    def parse_instance(self, instance, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def parse_instances(self, instances, **kwargs):
        raise NotImplementedError()



import logging
from typing import List

import tensorflow as tf
from naivenlp_datasets.masked_lm import (
    CsvFileReaderForMaskedLanguageModel,
    ExampleForMaskedLanguageModel,
    ExampleParserForMaskedLanguageModel,
    JsonlFileReaderForMaskedLanguageModel,
    TextFileReaderForMaskedLanguageModel,
)

from . import utils
from .dataset import TFDataset


class TFDatasetForMaksedLanguageModel(TFDataset):
    """Dataset for masked lm in TensorFlow"""

    def save_tfrecord(self, output_files, **kwargs):
        """Save examples in tfrecord format"""

        def _encode(example: ExampleForMaskedLanguageModel):
            feature = {
                "input_ids": utils.int64_feature([int(x) for x in example.input_ids]),
                "segment_ids": utils.int64_feature([int(x) for x in example.segment_ids]),
                "attention_mask": utils.int64_feature([int(x) for x in example.attention_mask]),
                "masked_ids": utils.int64_feature([int(x) for x in example.masked_ids]),
                "masked_pos": utils.int64_feature([int(x) for x in example.masked_pos]),
            }
            return feature

        utils.save_tfrecord(self.examples, _encode, output_files, **kwargs)

    @classmethod
    def from_tfrecord_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        dataset = utils.read_tfrecord_files(input_files, **kwargs)
        # parse example
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "masked_ids": tf.io.VarLenFeature(tf.int64),
            "masked_pos": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["masked_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["masked_pos"]), tf.int32),
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, **kwargs) -> tf.data.Dataset:
        reader = JsonlFileReaderForMaskedLanguageModel()
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_text_files(cls, input_files, vocab_file, **kwargs) -> tf.data.Dataset:
        reader = TextFileReaderForMaskedLanguageModel()
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_csv_files(cls, input_files, vocab_file, **kwargs):
        reader = CsvFileReaderForMaskedLanguageModel()
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_instances(cls, instances, vocab_file, **kwargs) -> tf.data.Dataset:
        parser = ExampleParserForMaskedLanguageModel(vocab_file, **kwargs)
        examples = []
        for e in parser.parse_instances(instances, **kwargs):
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls,
        examples: List[ExampleForMaskedLanguageModel],
        max_sequence_length=512,
        return_self=False,
        verbose=True,
        **kwargs
    ) -> tf.data.Dataset:
        "Parse examples to tf.data.Dataset"
        examples = [e for e in examples if len(e.input_ids) <= max_sequence_length]
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None
        if verbose:
            n = min(5, len(examples))
            for i in range(n):
                logging.info("Showing NO.%d example: %s", i, examples[i])

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        # conver examples to dataset
        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.masked_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.masked_pos for e in examples], dtype=tf.int32),
            )
        )
        # do transformation
        d = cls(examples=examples, **kwargs)
        if return_self:
            return d(dataset, **kwargs), d
        return d(dataset, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        if not do_filter:
            return dataset
        dataset = dataset.filter(lambda a, b, c, x, y: tf.size(a) <= max_sequence_length)
        return dataset

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        if not to_dict:
            dataset = dataset.map(
                lambda a, b, c, x, y: ((a, b, c), (x, y)),
                num_parallel_calls=num_parallel_calls,
            ).prefetch(buffer_size)
            return dataset
        dataset = dataset.map(
            lambda a, b, c, x, y: (
                {"input_ids": a, "segment_ids": b, "attention_mask": c},
                {"masked_ids": x, "masked_pos": y},
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        return dataset

    def _fixed_padding(self, dataset: tf.data.Dataset, pad_id=0, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen]))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, pad_id, pad_id))
        # fmt: on
        dataset = utils.batching_and_padding(dataset, padded_shapes, padding_values, **kwargs)
        return dataset

    def _batch_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([None, ], [None, ], [None, ], [None, ], [None, ]))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, pad_id, pad_id))
        # fmt: on
        dataset = utils.batching_and_padding(dataset, padded_shapes, padding_values, **kwargs)
        return dataset

    def _bucket_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([None, ], [None, ], [None, ], [None, ], [None, ]))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, pad_id, pad_id))
        # fmt: on
        dataset = utils.bucketing_and_padding(
            dataset,
            bucket_fn=lambda a, b, c, x, y: tf.size(a),
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            **kwargs,
        )
        return dataset



import unittest


class DatasetTest(unittest.TestCase):
    """Dataset test"""

    def test_dataset_tf(self):
        from naivenlp_datasets.tf import TFDatasetForMaksedLanguageModel

        dataset, d = TFDatasetForMaksedLanguageModel.from_jsonl_files(
            input_files=["testdata/mlm.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
            batch_size=4,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch:\n{}".format(idx, batch))

        d.save_tfrecord("testdata/mlm.tfrecord")

        dataset = TFDatasetForMaksedLanguageModel.from_tfrecord_files(
            input_files="testdata/mlm.tfrecord",
            batch_size=4,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch:\n{}".format(idx, batch))

    def test_dataset_pt(self):
        import torch
        from naivenlp_datasets.pt import DatasetForMaskedLanguageModel

        dataset = DatasetForMaskedLanguageModel.from_jsonl_files(
            input_files=["testdata/mlm.jsonl"],
            vocab_file="testdata/vocab.txt",
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, batch_size=4, num_workers=1, collate_fn=dataset.batch_padding_collate
        )
        for idx, batch in enumerate(dataloader):
            print()
            print("No.{} batch:\n{}".format(idx, batch))


if __name__ == "__main__":
    unittest.main()



import unittest


class DatasetForSimCSETest(unittest.TestCase):
    """SimCSE dataset tests"""

    def test_unsup_simcse_tf(self):
        from naivenlp_datasets.tf import TFDatasetForSimCSE, TFDatasetForUnsupSimCSE

        dataset, d = TFDatasetForUnsupSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        d.save_tfrecord("testdata/simcse_unsup.tfrecord")

        dataset = TFDatasetForUnsupSimCSE.from_tfrecord_files(
            input_files=["testdata/simcse_unsup.tfrecord"],
            vocab_file="testdata/vocab.txt",
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        # v2
        dataset, d = TFDatasetForSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
            with_pos_sequence=False,
            with_neg_sequence=False,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        d.save_tfrecord("testdata/simcse_unsup_v2.tfrecord")

        dataset = TFDatasetForSimCSE.from_tfrecord_files(
            input_files=["testdata/simcse_unsup_v2.tfrecord"],
            vocab_file="testdata/vocab.txt",
            with_pos_sequence=False,
            with_neg_sequence=False,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

    def test_sup_simcse_tf(self):
        from naivenlp_datasets.tf import TFDatasetForSimCSE, TFDatasetForSupervisedSimCSE

        dataset, d = TFDatasetForSupervisedSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        d.save_tfrecord("testdata/simcse_sup.tfrecord")

        dataset = TFDatasetForSupervisedSimCSE.from_tfrecord_files(
            input_files=["testdata/simcse_sup.tfrecord"],
            vocab_file="testdata/vocab.txt",
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        # v2
        dataset, d = TFDatasetForSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
            with_pos_sequence=True,
            with_neg_sequence=False,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        d.save_tfrecord("testdata/simcse_sup_v2.tfrecord")

        dataset = TFDatasetForSimCSE.from_tfrecord_files(
            input_files=["testdata/simcse_sup_v2.tfrecord"],
            vocab_file="testdata/vocab.txt",
            with_pos_sequence=True,
            with_neg_sequence=False,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

    def test_hardneg_simcse_tf(self):
        from naivenlp_datasets.tf import TFDatasetForHardNegativeSimCSE, TFDatasetForSimCSE

        dataset, d = TFDatasetForHardNegativeSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        d.save_tfrecord("testdata/simcse_hardneg.tfrecord")

        dataset = TFDatasetForHardNegativeSimCSE.from_tfrecord_files(
            input_files=["testdata/simcse_hardneg.tfrecord"],
            vocab_file="testdata/vocab.txt",
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        # v2
        dataset, d = TFDatasetForSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
            with_pos_sequence=True,
            with_neg_sequence=True,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

        d.save_tfrecord("testdata/simcse_hardneg_v2.tfrecord")

        dataset = TFDatasetForSimCSE.from_tfrecord_files(
            input_files=["testdata/simcse_hardneg_v2.tfrecord"],
            vocab_file="testdata/vocab.txt",
            with_pos_sequence=True,
            with_neg_sequence=True,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch: {}".format(idx, batch))

    def test_simcse_pt(self):
        import torch
        from naivenlp_datasets.pt import DatasetForSimCSE

        # unsup
        dataset = DatasetForSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            with_pos_sequence=False,
            with_neg_sequence=False,
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, num_workers=1, batch_size=4, collate_fn=dataset.batch_padding_collate
        )
        for idx, batch in enumerate(dataloader):
            print()
            print("No.{} batch: {}".format(idx, batch))

        # sup
        dataset = DatasetForSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            with_pos_sequence=True,
            with_neg_sequence=False,
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, num_workers=1, batch_size=4, collate_fn=dataset.batch_padding_collate
        )
        for idx, batch in enumerate(dataloader):
            print()
            print("No.{} batch: {}".format(idx, batch))

        # hardneg
        dataset = DatasetForSimCSE.from_jsonl_files(
            input_files=["testdata/simcse.jsonl"],
            vocab_file="testdata/vocab.txt",
            with_pos_sequence=True,
            with_neg_sequence=True,
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, num_workers=1, batch_size=4, collate_fn=dataset.batch_padding_collate
        )
        for idx, batch in enumerate(dataloader):
            print()
            print("No.{} batch: {}".format(idx, batch))


if __name__ == "__main__":
    unittest.main()



import abc
from collections import namedtuple

from tokenizers import BertWordPieceTokenizer

CharLevelEncoding = namedtuple("CharLevelEncoding", ["tokens", "ids", "type_ids", "attention_mask", "offsets"])


class CharLevelTokenizer(abc.ABC):
    """Abstract char-level tokenizer"""

    @abc.abstractmethod
    def encode(self, sequence, pair=None, is_pretokenized=False, add_special_tokens=True):
        raise NotImplementedError()

    @abc.abstractmethod
    def encode_batch(self, inputs, is_pretokenized=False, add_special_tokens=True):
        raise NotImplementedError()

    @abc.abstractmethod
    def decode(self, ids, skip_special_tokens=True):
        raise NotImplementedError()

    @abc.abstractmethod
    def decode_batch(self, sequences, skip_special_tokens=True):
        raise NotImplementedError()


class BertCharLevelTokenizer(CharLevelTokenizer):
    """Bert char-level tokenizer"""

    @classmethod
    def from_file(cls, vocab_file, lowercase=True, **kwargs):
        tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=lowercase)
        return cls(tokenizer=tokenizer, **kwargs)

    def __init__(self, tokenizer: BertWordPieceTokenizer, **kwargs) -> None:
        super().__init__()
        self.tokenizer = tokenizer

    @property
    def pad_token(self):
        return self.tokenizer._parameters["cls_token"]

    @property
    def unk_token(self):
        return self.tokenizer._parameters["unk_token"]

    @property
    def cls_token(self):
        return self.tokenizer._parameters["cls_token"]

    @property
    def sep_token(self):
        return self.tokenizer._parameters["sep_token"]

    @property
    def mask_token(self):
        return self.tokenizer._parameters["mask_token"]

    @property
    def special_tokens(self):
        return [self.unk_token, self.pad_token, self.cls_token, self.sep_token, self.mask_token]

    def token_to_id(self, token):
        return self.tokenizer.token_to_id(token)

    def id_to_token(self, _id):
        return self.tokenizer.id_to_token(_id)

    def _to_char_level(self, encoding, **kwargs):
        tokens, ids, type_ids, attention_mask, offsets = [], [], [], [], []
        for token, _id, tid, mask, offset in zip(
            encoding.tokens, encoding.ids, encoding.type_ids, encoding.attention_mask, encoding.offsets
        ):
            if token in self.special_tokens:
                tokens.append(token)
                ids.append(_id)
                type_ids.append(tid)
                attention_mask.append(mask)
                offsets.append(offset)
                continue
            token = str(token).lstrip("##")
            for idx, c in enumerate(token):
                ids.append(len(tokens))
                tokens.append(c)
                type_ids.append(tid)
                attention_mask.append(mask)
                start = offset[0]
                offsets.append((start + idx, start + idx + 1))
        char_encoding = CharLevelEncoding(
            tokens=tokens,
            ids=ids,
            type_ids=type_ids,
            attention_mask=attention_mask,
            offsets=offsets,
        )
        return char_encoding

    def encode(self, sequence, pair=None, is_pretokenized=False, add_special_tokens=True):
        encoding = self.tokenizer.encode(
            sequence, pair=pair, is_pretokenized=is_pretokenized, add_special_tokens=add_special_tokens
        )
        return self._to_char_level(encoding)

    def encode_batch(self, inputs, is_pretokenized=False, add_special_tokens=True):
        encodings = self.tokenizer.encode_batch(
            inputs, is_pretokenized=is_pretokenized, add_special_tokens=add_special_tokens
        )
        return [self._to_char_level(encoding) for encoding in encodings]

    def decode(self, ids, skip_special_tokens=True):
        return self.tokenizer.decode(ids, skip_special_tokens=skip_special_tokens)

    def decode_batch(self, sequences, skip_special_tokens=True):
        return self.tokenizer.decode_batch(sequences, skip_special_tokens=skip_special_tokens)



import abc
import logging
import os
import random
import re
from collections import namedtuple

from tokenizers import BertWordPieceTokenizer

from .parsers import AbstractExampleParser
from .readers import AbstractFileReader, CsvFileReader, JsonlFileReader

ExampleForMaskedLanguageModel = namedtuple(
    "ExampleForMaskedLanguageModel",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "masked_ids", "masked_pos"],
)


ResultForMasking = namedtuple("ResultForMasking", ["origin_tokens", "masked_tokens", "masked_indexes"])


class AbstractMaskingStrategy(abc.ABC):
    """Abstract masking strategy"""

    @abc.abstractmethod
    def __call__(self, tokens, **kwargs) -> ResultForMasking:
        raise NotImplementedError()


class WholeWordMask(AbstractMaskingStrategy):
    """Default masking strategy from BERT."""

    def __init__(
        self, vocabs, change_prob=0.15, mask_prob=0.8, rand_prob=0.1, keep_prob=0.1, max_predictions=20, **kwargs
    ):
        self.vocabs = vocabs
        self.change_prob = change_prob
        self.mask_prob = mask_prob / (mask_prob + rand_prob + keep_prob)
        self.rand_prob = rand_prob / (mask_prob + rand_prob + keep_prob)
        self.keep_prob = keep_prob / (mask_prob + rand_prob + keep_prob)
        self.max_predictions = max_predictions

    def __call__(self, tokens, max_sequence_length=512, **kwargs) -> ResultForMasking:
        tokens = self._truncate_sequence(tokens, max_sequence_length - 2)
        if not tokens:
            return None
        num_to_predict = min(self.max_predictions, max(1, round(self.change_prob * len(tokens))))
        cand_indexes = self._collect_candidates(tokens)
        # copy original tokens
        masked_tokens = [x for x in tokens]
        masked_indexes = [0] * len(tokens)
        for piece_indexes in cand_indexes:
            if sum(masked_indexes) >= num_to_predict:
                break
            if sum(masked_indexes) + len(piece_indexes) > num_to_predict:
                continue
            if any(masked_indexes[idx] == 1 for idx in piece_indexes):
                continue
            for index in piece_indexes:
                masked_indexes[index] = 1
                masked_tokens[index] = self._masking_tokens(index, tokens, self.vocabs)

        # add special tokens
        tokens = ["[CLS]"] + tokens + ["[SEP]"]
        masked_tokens = ["[CLS]"] + masked_tokens + ["[SEP]"]
        masked_indexes = [0] + masked_indexes + [0]
        assert len(tokens) == len(masked_tokens) == len(masked_indexes)
        return ResultForMasking(origin_tokens=tokens, masked_tokens=masked_tokens, masked_indexes=masked_indexes)

    def _masking_tokens(self, index, tokens, vocabs, **kwargs):
        # 80% of the time, replace with [MASK]
        if random.random() < self.mask_prob:
            return "[MASK]"
        # 10% of the time, keep original
        p = self.rand_prob / (self.rand_prob + self.keep_prob)
        if random.random() < p:
            return tokens[index]
        # 10% of the time, replace with random word
        masked_token = vocabs[random.randint(0, len(vocabs) - 1)]
        return masked_token

    def _collect_candidates(self, tokens):
        cand_indexes = [[]]
        for idx, token in enumerate(tokens):
            if cand_indexes and token.startswith("##"):
                cand_indexes[-1].append(idx)
                continue
            cand_indexes.append([idx])
        random.shuffle(cand_indexes)
        return cand_indexes

    def _truncate_sequence(self, tokens, max_tokens=512, **kwargs):
        while len(tokens) > max_tokens:
            if len(tokens) > max_tokens:
                tokens.pop(0)
                # truncate whole world
                while tokens and tokens[0].startswith("##"):
                    tokens.pop(0)
            if len(tokens) > max_tokens:
                while tokens and tokens[-1].startswith("##"):
                    tokens.pop()
                if tokens:
                    tokens.pop()
        return tokens


class JsonlFileReaderForMaskedLanguageModel(JsonlFileReader):
    """Jsonl file reader for mlm"""

    def _parse_instance(self, data, sequence_column="sequence", **kwargs):
        return {"sequence": data[sequence_column]}


class CsvFileReaderForMaskedLanguageModel(CsvFileReader):
    """Csv file reader for mlm"""

    def _parse_instance(self, line, sequence_column=0, sep=",", **kwargs):
        parts = re.split(sep, line)
        return {"sequence": parts[sequence_column]}


class TextFileReaderForMaskedLanguageModel(AbstractFileReader):
    """Text file reader for mlm"""

    def read_files(self, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        for f in input_files:
            if not os.path.exists(f) or not os.path.isfile(f):
                logging.warning("File %d does not exist, skipped.", f)
                continue
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    instance = {"sequence": line}
                    yield instance


class ExampleParserForMaskedLanguageModel(AbstractExampleParser):
    """Parse instance to ExampleForMaskedLanguageModel"""

    def __init__(self, vocab_file, masking="whole-word-mask", do_lower_case=True, **kwargs) -> None:
        super().__init__()
        self.tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)
        vocabs = list(self.tokenizer.get_vocab().keys())
        assert masking in ["whole-word-mask"]
        if masking == "whole-word-mask":
            self.masking = WholeWordMask(vocabs=vocabs, **kwargs)
        else:
            self.masking = None

    def parse_instance(self, instance, max_sequence_length=512, **kwargs):
        sequence = instance["sequence"]
        encoding = self.tokenizer.encode(sequence, add_special_tokens=False)
        masking_results = self.masking(tokens=encoding.tokens, max_sequence_length=max_sequence_length, **kwargs)
        origin_tokens, masked_tokens = masking_results.origin_tokens, masking_results.masked_tokens
        example = ExampleForMaskedLanguageModel(
            tokens=masked_tokens,
            input_ids=[self.tokenizer.token_to_id(x) for x in masked_tokens],
            segment_ids=[0] * len(masked_tokens),
            attention_mask=[1] * len(masked_tokens),
            masked_ids=[self.tokenizer.token_to_id(x) for x in origin_tokens],
            masked_pos=masking_results.masked_indexes,
        )
        return example

    def parse_instances(self, instances, max_sequence_length=512, **kwargs):
        for instance in instances:
            e = self.parse_instance(instance, max_sequence_length=max_sequence_length, **kwargs)
            if not e:
                continue
            yield e



import unittest


class DatasetForSequenceClassificationTest(unittest.TestCase):
    """Dataset for sequence classification"""

    def test_dataset_for_sequence_classification_pt(self):
        import torch
        from naivenlp_datasets.pt import DatasetForSequenceClassification

        dataset = DatasetForSequenceClassification.from_jsonl_files(
            input_files=["testdata/sequence_classification.jsonl"],
            vocab_file="testdata/vocab.txt",
        )
        for idx, batch in enumerate(
            torch.utils.data.DataLoader(
                dataset, num_workers=2, shuffle=True, batch_size=32, collate_fn=dataset.batch_padding_collate
            )
        ):
            print()
            print("NO.{} batch: \n{}".format(idx, batch))

    def test_dataset_for_sequence_classification_tf(self):
        from naivenlp_datasets.tf import TFDatasetForSequenceClassifiation

        dataset, d = TFDatasetForSequenceClassifiation.from_jsonl_files(
            input_files=["testdata/sequence_classification.jsonl"],
            vocab_file="testdata/vocab.txt",
            return_self=True,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("NO.{} batch: \n{}".format(idx, batch))

        print("Size of examples: ", len(d.examples))
        d.save_tfrecord("testdata/sequence_classification.tfrecord")

        dataset = TFDatasetForSequenceClassifiation.from_tfrecord_files("testdata/sequence_classification.tfrecord")
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("NO.{} batch: \n{}".format(idx, batch))


if __name__ == "__main__":
    unittest.main()



import re

from tokenizers import BertWordPieceTokenizer

from .parsers import AbstractExampleParser
from .readers import CsvFileReader, JsonlFileReader


class ExampleForSimCSE:
    """Example for SimCSE model"""

    def __init__(self, **kwargs) -> None:
        self.tokens = kwargs.get("tokens", None)
        self.input_ids = kwargs.get("input_ids", None)
        self.segment_ids = kwargs.get("segment_ids", None)
        self.attention_mask = kwargs.get("attention_mask", None)
        self.pos_tokens = kwargs.get("pos_tokens", None)
        self.pos_input_ids = kwargs.get("pos_input_ids", None)
        self.pos_segment_ids = kwargs.get("pos_segment_ids", None)
        self.pos_attention_mask = kwargs.get("pos_attention_mask", None)
        self.neg_tokens = kwargs.get("neg_tokens", None)
        self.neg_input_ids = kwargs.get("neg_input_ids", None)
        self.neg_segment_ids = kwargs.get("neg_segment_ids", None)
        self.neg_attention_mask = kwargs.get("neg_attention_mask", None)

    def __hash__(self) -> int:
        values = [getattr(self, k, None) for k in self.__dict__.keys()]
        return hash(values)

    def __eq__(self, __o: object) -> bool:
        if not __o:
            return False
        if not isinstance(__o, ExampleForSimCSE):
            return False
        for k, v in self.__dict__.items():
            if v != getattr(__o, k, None):
                return False
        return True

    def __repr__(self) -> str:
        builder = []
        for k, v in self.__dict__.items():
            builder.append("{}={}".format(k, v))
        return "ExampleForSimCSE[" + ",".join(builder) + "]"

    def __str__(self) -> str:
        return self.__repr__()


class JsonlFileReaderForSimCSE(JsonlFileReader):
    """Jsonl file reader for simcse"""

    def __init__(
        self,
        sequence_column="sequence",
        with_pos_sequence=False,
        pos_sequence_column="pos_sequence",
        with_neg_sequence=False,
        neg_sequence_column="neg_sequence",
        **kwargs
    ) -> None:
        super().__init__()
        self.sequence_column = sequence_column
        self.with_pos_sequence = with_pos_sequence
        self.pos_sequence_column = pos_sequence_column
        self.with_neg_sequence = with_neg_sequence
        self.neg_sequence_column = neg_sequence_column

    def _parse_instance(self, data, **kwargs):
        instance = {
            "sequence": data[self.sequence_column],
        }
        if self.with_pos_sequence and self.pos_sequence_column in data:
            instance["pos_sequence"] = data[self.pos_sequence_column]
        if self.with_neg_sequence and self.neg_sequence_column in data:
            instance["neg_sequence"] = data[self.neg_sequence_column]
        return instance


class CsvFileReaderForSimCSE(CsvFileReader):
    """Csv file reader for simcse"""

    def __init__(
        self,
        sequence_column=0,
        with_pos_sequence=False,
        pos_sequence_column=1,
        with_neg_sequence=False,
        neg_sequence_column=2,
        **kwargs
    ) -> None:
        super().__init__()
        self.sequence_column = sequence_column
        self.with_pos_sequence = with_pos_sequence
        self.pos_sequence_column = pos_sequence_column
        self.with_neg_sequence = with_neg_sequence
        self.neg_sequence_column = neg_sequence_column

    def _parse_instance(self, line, sep=",", **kwargs):
        data = re.split(sep, line)
        instance = {
            "sequence": data[self.sequence_column],
        }
        if self.with_pos_sequence and self.pos_sequence_column in data:
            instance["pos_sequence"] = data[self.pos_sequence_column]
        if self.with_neg_sequence and self.neg_sequence_column in data:
            instance["neg_sequence"] = data[self.neg_sequence_column]
        return instance


class ExampleParserForSimCSE(AbstractExampleParser):
    """Parse json instance to ExampleForSimCSE"""

    def __init__(
        self, vocab_file, with_pos_sequence=False, with_neg_sequence=False, do_lower_case=True, **kwargs
    ) -> None:
        super().__init__()
        self.tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)
        self.with_pos_sequence = with_pos_sequence
        self.with_neg_sequence = with_neg_sequence

    def parse_instance(self, instance, add_special_tokens=True, **kwargs):
        seq_encoding = self.tokenizer.encode(instance["sequence"], add_special_tokens=add_special_tokens)
        e = ExampleForSimCSE(
            tokens=seq_encoding.tokens,
            input_ids=seq_encoding.ids,
            segment_ids=seq_encoding.type_ids,
            attention_mask=seq_encoding.attention_mask,
        )
        if self.with_pos_sequence:
            pos_encoding = self.tokenizer.encode(instance["pos_sequence"], add_special_tokens=add_special_tokens)
            e.pos_tokens = pos_encoding.tokens
            e.pos_input_ids = pos_encoding.ids
            e.pos_segment_ids = pos_encoding.type_ids
            e.pos_attention_mask = pos_encoding.attention_mask
        if self.with_neg_sequence:
            neg_encoding = self.tokenizer.encode(instance["neg_sequence"], add_special_tokens=add_special_tokens)
            e.neg_tokens = neg_encoding.tokens
            e.neg_input_ids = neg_encoding.ids
            e.neg_segment_ids = neg_encoding.type_ids
            e.neg_attention_mask = neg_encoding.attention_mask
        return e

    def parse_instances(self, instances, add_special_tokens=True, **kwargs):
        for instance in instances:
            e = self.parse_instance(instance, add_special_tokens=add_special_tokens, **kwargs)
            yield e



import abc

import torch


class PTDataset(torch.utils.data.Dataset):
    """Dataset in PyTorch"""

    def __init__(self, pad_id=0, max_sequence_length=512, **kwargs) -> None:
        super().__init__()
        self.pad_id = pad_id
        self.max_sequence_length = max_sequence_length

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def __getitem__(self, index):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def batch_padding_collate(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def fixed_padding_collate(self):
        raise NotImplementedError()



from .masked_lm_dataset import DatasetForMaskedLanguageModel
from .question_answering_dataset import DatasetForQuestionAnswering
from .sequence_classification_dataset import DatasetForSequenceClassification
from .simcse_dataset import DatasetForSimCSE
from .token_classification_dataset import DatasetForTokenClassification



import abc
import json
import logging
import os
import re


class AbstractFileReader(abc.ABC):
    """Abstract file reader"""

    @abc.abstractmethod
    def read_files(self, input_files, **kwargs):
        raise NotImplementedError()


class JsonlFileReader(AbstractFileReader):
    """Jsonl file reader"""

    def __init__(self) -> None:
        super().__init__()

    def read_files(self, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        for f in input_files:
            if not os.path.exists(f) or not os.path.isfile(f):
                logging.warning("File %d does not exist, skipped.", f)
                continue
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    data = json.loads(line)
                    instance = self._parse_instance(data, **kwargs)
                    if not instance:
                        continue
                    yield instance

    @abc.abstractmethod
    def _parse_instance(self, data, **kwargs):
        raise NotImplementedError()


class CsvFileReader(AbstractFileReader):
    """CSV file reader"""

    def __init__(self) -> None:
        super().__init__()

    def read_files(self, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        for f in input_files:
            if not os.path.exists(f) or not os.path.isfile(f):
                logging.warning("File %d does not exist, skipped.", f)
                continue
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    instance = self._parse_instance(line, **kwargs)
                    if not instance:
                        continue
                    yield instance

    @abc.abstractmethod
    def _parse_instance(self, line, sep=",", **kwargs):
        raise NotImplementedError()


class ConllFileReader(AbstractFileReader):
    """CONLL file reader"""

    def read_files(self, input_files, feature_column=0, label_column=1, sep="\t", **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        for f in input_files:
            if not os.path.exists(f) or not os.path.isfile(f):
                logging.warning("File %d does not exist, skipped.", f)
                continue
            feature, label = []
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        instance = self._parse_instance(feature, label, **kwargs)
                        yield instance
                        feature, label = [], []
                        continue
                    parts = re.split(sep, line)
                    feature.append(parts[feature_column])
                    label.append(parts[label_column])
            if feature and label:
                instance = self._parse_instance(feature, label, **kwargs)
                yield instance

    @abc.abstractmethod
    def _parse_instance(self, feature, label, **kwargs):
        raise NotImplementedError()



import unittest

from naivenlp_datasets.charlevel_tokenizer import BertCharLevelTokenizer


class CharLevelTokenizerTest(unittest.TestCase):
    """Char level tokenizer test"""

    def test_bert_char_level_tokenizer(self):
        print()
        tokenizer = BertCharLevelTokenizer.from_file("testdata/vocab.txt", lowercase=True)
        encoding = tokenizer.encode("我喜欢自然语言处理NLP! It's amazing!")
        # fmt: off
        self.assertEqual(encoding.tokens, ['[CLS]', '我', '喜', '欢', '自', '然', '语', '言', '处', '理', 'n', 'l', 'p', '!', 'i', 't', "'", 's', 'a', 'm', 'a', 'z', 'i', 'n', 'g', '!', '[SEP]'])
        # fmt: on

        encoding = tokenizer.encode("我喜欢自然语言处理NLP! ", "It's amazing!")
        # fmt: off
        self.assertEqual(encoding.tokens, ['[CLS]', '我', '喜', '欢', '自', '然', '语', '言', '处', '理', 'n', 'l', 'p', '!', '[SEP]', 'i', 't', "'", 's', 'a', 'm', 'a', 'z', 'i', 'n', 'g', '!', '[SEP]'])
        # fmt: on

        encodings = tokenizer.encode_batch(["我喜欢自然语言处理NLP! It's amazing!"] * 4, add_special_tokens=False)
        for encoding in encodings:
            # fmt: off
            self.assertEqual(encoding.tokens, ['我', '喜', '欢', '自', '然', '语', '言', '处', '理', 'n', 'l', 'p', '!', 'i', 't', "'", 's', 'a', 'm', 'a', 'z', 'i', 'n', 'g', '!'])
            # fmt: on


if __name__ == "__main__":
    unittest.main()



from typing import List

from naivenlp_datasets.token_classification import (
    ConllFileReaderForTokenClassification,
    ExampleForTokenClassification,
    ExampleParserForTokenClassification,
    JsonlFileReaderForTokenClassification,
)

from . import utils
from .dataset import PTDataset


class DatasetForTokenClassification(PTDataset):
    """Dataset for token classification in PyTorch"""

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, label_to_id, tokenization="bert-wordpiece", **kwargs):
        reader = JsonlFileReaderForTokenClassification()
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, label_to_id, tokenization=tokenization, **kwargs)

    @classmethod
    def from_conll_files(cls, input_files, vocab_file, label_to_id, tokenization="bert-wordpiece", **kwargs):
        reader = ConllFileReaderForTokenClassification()
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, label_to_id, tokenization=tokenization, **kwargs)

    @classmethod
    def from_instances(cls, instances, vocab_file, label_to_id, tokenization="bert-wordpiece", **kwargs):
        parser = ExampleParserForTokenClassification(vocab_file, label_to_id, tokenization=tokenization, **kwargs)
        examples = []
        for e in parser.parse_instances(instances, **kwargs):
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(cls, examples: List[ExampleForTokenClassification], max_sequence_length=512, **kwargs):
        examples = [e for e in examples if len(e.input_ids) <= max_sequence_length]
        return cls(examples=examples, max_sequence_length=max_sequence_length, **kwargs)

    def __init__(self, examples: List[ExampleForTokenClassification], **kwargs) -> None:
        super().__init__(**kwargs)
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        e = self.examples[index]
        features = {
            "input_ids": e.input_ids,
            "segment_ids": e.segment_ids,
            "attention_mask": e.attention_mask,
        }
        labels = {
            "label_ids": e.label_ids,
        }
        return features, labels

    @property
    def batch_padding_collate(self):
        """Padding sequence to max length in a batch"""

        def _collate_fn(batch):
            max_length = max([len(x["input_ids"]) for x, _ in batch])
            padded_features = utils.padding_features(
                features=[x for x, _ in batch], pad_id=self.pad_id, max_length=max_length
            )
            padded_labels = utils.padding_labels(
                labels=[y for _, y in batch], pad_id=self.pad_id, max_length=max_length
            )
            return padded_features, padded_labels

        return _collate_fn

    @property
    def fixed_padding_collate(self):
        """Padding sequence to a fixed max length"""

        def _collate_fn(batch):
            max_length = self.max_sequence_length
            padded_features = utils.padding_features(
                features=[x for x, _ in batch], pad_id=self.pad_id, max_length=max_length
            )
            padded_labels = utils.padding_labels(
                labels=[y for _, y in batch], pad_id=self.pad_id, max_length=max_length
            )
            return padded_features, padded_labels

        return _collate_fn



import json
import logging
import re
from collections import namedtuple
from typing import Iterator, List

from tokenizers import BertWordPieceTokenizer

from naivenlp_datasets.readers import JsonlFileReader

from .charlevel_tokenizer import BertCharLevelTokenizer

ExampleForQuestionAnswering = namedtuple(
    "ExampleForQuestionAnswering",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "start_positions", "end_positions"],
)


class JsonlFileReaderForQuestionAnswering(JsonlFileReader):
    """Jsonl file reader for question answering"""

    def _parse_instance(
        self, data, context_column="context", question_column="question", answer_column="answer", **kwargs
    ):
        instance = {
            "context": data[context_column],
            "question": data[question_column],
        }
        answer = data[answer_column]
        # if has multi answers, use the first one
        if isinstance(answer, (List, list)):
            answer = answer[0]
        instance["answer"] = answer
        return instance


class ExampleParserForQuestionAnswering:
    """Parse json instance to ExampleForQuestionAnswering"""

    def __init__(self, vocab_file, tokenization="bert-wordpiece", do_lower_case=True, **kwargs) -> None:
        assert tokenization in ["bert-wordpiece", "bert-charlevel"], "Unsupported tokenization: {}".format(tokenization)
        if tokenization == "bert-wordpiece":
            self.tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)
        if tokenization == "bert-charlevel":
            self.tokenizer = BertCharLevelTokenizer.from_file(vocab_file, lowercase=do_lower_case)

    @property
    def sep_token(self):
        if isinstance(self.tokenizer, BertCharLevelTokenizer):
            return self.tokenizer.sep_token
        if isinstance(self.tokenizer, BertWordPieceTokenizer):
            return self.tokenizer._parameters["sep_token"]
        raise ValueError("Invalid type of self.tokenizer.")

    def _find_answer_span(self, context, answer):
        for m in re.finditer(re.escape(answer), context, re.IGNORECASE):
            start, end = m.span()
            return start, end
        return 0, 0

    def _parse(self, context, question, answer, answer_start, answer_end, **kwargs):
        encoding = self.tokenizer.encode(context, question, add_special_tokens=True)
        is_char_in_ans = [0] * len(context)
        for idx in range(answer_start, answer_end):
            is_char_in_ans[idx] = 1
        ans_token_index = []
        for idx, (start_char_index, end_char_index) in enumerate(encoding.offsets):
            if encoding.tokens[idx] == self.sep_token:
                break
            if sum(is_char_in_ans[start_char_index:end_char_index]) > 0:
                ans_token_index.append(idx)
        if not ans_token_index:
            logging.debug("No answering found. Skipped.")
            return None

        ans_token_start, ans_token_end = ans_token_index[0], ans_token_index[-1]
        valid = self._valid_example(answer, encoding.tokens, ans_token_start, ans_token_end + 1)
        if not valid:
            return None
        return ExampleForQuestionAnswering(
            tokens=encoding.tokens,
            input_ids=encoding.ids,
            segment_ids=encoding.type_ids,
            attention_mask=encoding.attention_mask,
            start_positions=ans_token_start,
            end_positions=ans_token_end,
        )

    def _valid_example(self, answer, tokens, start, end):
        sub_tokens = tokens[start:end]
        ans_tokens = self.tokenizer.encode(answer, add_special_tokens=False).tokens
        if ans_tokens != sub_tokens:
            logging.debug("Invalid example. ans_tokens: %s\t sub_tokens: %s", ans_tokens, sub_tokens)
            return False
        return True

    def parse_instance(
        self,
        instance,
        context_column="context",
        question_column="question",
        answer_column="answer",
        answer_start_column=None,
        **kwargs
    ) -> ExampleForQuestionAnswering:
        context, question, answer = instance[context_column], instance[question_column], instance[answer_column]
        if answer_start_column is not None and answer_start_column in instance:
            answer_start = int(instance[answer_start_column])
            answer_end = answer_start + len(answer)
        else:
            answer_start, answer_end = self._find_answer_span(context, answer)

        if answer_end <= answer_start:
            logging.warning("Invalid answer span, skipped.")
            return None
        return self._parse(context, question, answer, answer_start, answer_end, **kwargs)

    def parse_instances(
        self,
        instances,
        context_column="context",
        question_column="question",
        answer_column="answer",
        answer_start_column=None,
        **kwargs
    ) -> Iterator[ExampleForQuestionAnswering]:
        for instance in instances:
            example = self.parse_instance(
                instance,
                context_column=context_column,
                question_column=question_column,
                answer_column=answer_column,
                answer_start_column=answer_start_column,
                **kwargs
            )
            if not example:
                continue
            yield example


def read_dureader_rubost(input_files, **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for input_file in input_files:
        with open(input_file, mode="rt", encoding="utf-8") as fin:
            info = json.load(fin)
        for d in info["data"]:
            paragraphs = d["paragraphs"]
            for p in paragraphs:
                context = p["context"]
                for qa in p["qas"]:
                    question = qa["question"]
                    _id = qa["id"]
                    if not qa["answers"]:
                        continue
                    answer = qa["answers"][0]["text"]
                    instance = {"context": context, "question": question, "answer": answer, "id": _id}
                    yield instance


def read_dureader_checklist(input_files, **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for input_file in input_files:
        with open(input_file, mode="rt", encoding="utf-8") as fin:
            info = json.load(fin)
        for data in info["data"]:
            for p in data["paragraphs"]:
                # title = p["title"]
                context = p["context"]
                for qa in p["qas"]:
                    if qa["is_impossible"]:
                        continue
                    question = qa["question"]
                    answer = qa["answers"][0]["text"]
                    instance = {"context": context, "question": question, "answer": answer, "id": qa["id"]}
                    yield instance



from typing import List

import torch
from naivenlp_datasets.question_answering import (
    ExampleForQuestionAnswering,
    ExampleParserForQuestionAnswering,
    JsonlFileReaderForQuestionAnswering,
    read_dureader_checklist,
    read_dureader_rubost,
)


def _padding_features(features, pad_id=0, max_length=512):
    padded_features = {}
    for feature in features:
        for k in feature.keys():
            if k not in padded_features:
                padded_features[k] = []
            value = feature[k]
            padded_value = value + [pad_id] * (max_length - len(value))
            padded_features[k].append(padded_value)
    # to tensor
    return {k: torch.LongTensor(v) for k, v in padded_features.items()}


def _padding_labels(labels):
    padded_labels = {
        "start_positions": torch.LongTensor([x["start_positions"] for x in labels]),
        "end_positions": torch.LongTensor([x["end_positions"] for x in labels]),
    }
    return padded_labels


class DatasetForQuestionAnswering(torch.utils.data.Dataset):
    """Dataset for question answering in PyTorch"""

    @classmethod
    def from_dureader_robust(cls, input_files, vocab_file, tokenization="bert-wordpiece", **kwargs):
        instances = read_dureader_rubost(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, tokenization=tokenization, **kwargs)

    @classmethod
    def from_dureader_checklist(cls, input_files, vocab_file, tokenization="bert-charlevel", **kwargs):
        instances = read_dureader_checklist(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, tokenization=tokenization, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, tokenization="bert-wordpiece", **kwargs):
        reader = JsonlFileReaderForQuestionAnswering()
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, tokenization=tokenization, **kwargs)

    @classmethod
    def from_instances(cls, instances, vocab_file, tokenization="bert-wordpiece", **kwargs):
        parser = ExampleParserForQuestionAnswering(vocab_file, tokenization=tokenization, **kwargs)
        examples = []
        for e in parser.parse_instances(instances, **kwargs):
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(cls, examples: List[ExampleForQuestionAnswering], pad_id=0, max_sequence_length=512, **kwargs):
        examples = [e for e in examples if len(e.input_ids) <= max_sequence_length]
        return cls(examples=examples, pad_id=pad_id, max_sequence_length=max_sequence_length, **kwargs)

    def __init__(
        self, examples: List[ExampleForQuestionAnswering], pad_id=0, max_sequence_length=512, **kwargs
    ) -> None:
        super().__init__()
        self.examples = examples
        self.pad_id = pad_id
        self.max_sequence_length = max_sequence_length

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        e = self.examples[index]
        features = {
            "input_ids": e.input_ids,
            "segment_ids": e.segment_ids,
            "attention_mask": e.attention_mask,
        }
        labels = {
            "start_positions": e.start_positions,
            "end_positions": e.end_positions,
        }
        return features, labels

    @property
    def batch_padding_collate(self):
        """Padding sequence to max length in batch"""

        def _collate_fn(batch):
            max_length = max([len(x["input_ids"]) for x, _ in batch])
            features = _padding_features([x for x, _ in batch], pad_id=self.pad_id, max_length=max_length)
            labels = _padding_labels([y for _, y in batch])
            return features, labels

        return _collate_fn

    @property
    def fixed_padding_collate(self):
        """Padding sequence to fixed length"""

        def _collate_fn(batch):
            max_length = self.max_sequence_length
            features = _padding_features([x for x, _ in batch], pad_id=self.pad_id, max_length=max_length)
            labels = _padding_labels([y for _, y in batch])
            return features, labels

        return _collate_fn



import os
import unittest

from naivenlp_datasets.question_answering import ExampleParserForQuestionAnswering, read_dureader_rubost


class ExampleParserTest(unittest.TestCase):
    """Example prser tests"""

    def test_wordpiece_parser(self):
        p = ExampleParserForQuestionAnswering("testdata/vocab.txt", tokenization="bert-wordpiece", do_lower_case=True)
        input_file = os.path.join(os.environ["DUREADER_ROBUST_PATH"], "dev.json")
        idx = 0
        for instance in read_dureader_rubost(input_file):
            e = p.parse_instance(instance)
            idx += 1
            if not e:
                print("index=", idx)

    def test_charlevel_parser(self):
        p = ExampleParserForQuestionAnswering("testdata/vocab.txt", tokenization="bert-charlevel", do_lower_case=True)
        input_file = os.path.join(os.environ["DUREADER_ROBUST_PATH"], "dev.json")
        idx = 0
        for instance in read_dureader_rubost(input_file):
            e = p.parse_instance(instance)
            idx += 1
            if not e:
                print("index=", idx)


class QuestionAnsweringDatasetTest(unittest.TestCase):
    """Question answering dataset test"""

    def test_question_answering_dataset_pt(self):
        import torch
        from naivenlp_datasets.pt import DatasetForQuestionAnswering

        dataset = DatasetForQuestionAnswering.from_dureader_robust(
            input_files=os.path.join(os.environ["DUREADER_ROBUST_PATH"], "dev.json"),
            vocab_file="testdata/vocab.txt",
        )
        for idx, batch in enumerate(
            torch.utils.data.DataLoader(
                dataset, num_workers=2, shuffle=True, batch_size=32, collate_fn=dataset.batch_padding_collate
            )
        ):
            print()
            print("NO.{} batch: \n{}".format(idx, batch))

    def test_question_answering_dataset_tf(self):
        from naivenlp_datasets.tf import TFDatasetForQuestionAnswering

        dataset, d = TFDatasetForQuestionAnswering.from_dureader_robust(
            input_files=os.path.join(os.environ["DUREADER_ROBUST_PATH"], "dev.json"),
            vocab_file="testdata/vocab.txt",
            return_self=True,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("NO.{} batch: \n{}".format(idx, batch))

        print("Size of examples: ", len(d.examples))
        d.save_tfrecord("testdata/dureader_robust.tfrecord")

        dataset = TFDatasetForQuestionAnswering.from_tfrecord_files(
            "testdata/dureader_robust.tfrecord",
            batch_size=32,
            padding="batch",
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("NO.{} batch: \n{}".format(idx, batch))


if __name__ == "__main__":
    unittest.main()



from .dataset import TFDataset
from .masked_lm_dataset import TFDatasetForMaksedLanguageModel
from .question_answering_dataset import TFDatasetForQuestionAnswering
from .sequence_classification_dataset import TFDatasetForSequenceClassifiation
from .simcse_dataset import (
    TFDatasetForHardNegativeSimCSE,
    TFDatasetForSimCSE,
    TFDatasetForSupervisedSimCSE,
    TFDatasetForUnsupSimCSE,
)
from .token_classification_dataset import TFDatasetForTokenClassification



from collections import namedtuple
from typing import Iterator

from tokenizers import BertWordPieceTokenizer

from .charlevel_tokenizer import BertCharLevelTokenizer
from .parsers import AbstractExampleParser
from .readers import ConllFileReader, CsvFileReader, JsonlFileReader

ExampleForTokenClassification = namedtuple(
    "ExampleForTokenClassification", ["tokens", "input_ids", "segment_ids", "attention_mask", "label_ids"]
)


class JsonlFileReaderForTokenClassification(JsonlFileReader):
    """Jsonl file reader for token classification"""

    def _parse_instance(self, data, feature_column="feature", label_column="label", **kwargs):
        instance = {
            "features": data[feature_column],
            "labels": data[label_column],
        }
        return instance


class ConllFileReaderForTokenClassification(ConllFileReader):
    """CONLL file reader for token classification"""

    def _parse_instance(self, feature, label, **kwargs):
        instance = {
            "features": feature,
            "labels": label,
        }
        return instance


class ExampleParserForTokenClassification(AbstractExampleParser):
    """Parse instance to ExampleForTokenClassification"""

    def __init__(self, vocab_file, label_to_id, tokenization="bert-charlevel", do_lower_case=True, **kwargs) -> None:
        super().__init__()
        self.label_to_id = label_to_id
        assert tokenization in ["bert-charlevel", "bert-wordpiece"]
        if tokenization == "bert-charlevel":
            self.tokenizer = BertCharLevelTokenizer.from_file(vocab_file, lowercase=do_lower_case)
        if tokenization == "bert-wordpiece":
            self.tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)

    def parse_instance(self, instance, **kwargs) -> ExampleForTokenClassification:
        tokens = instance["features"]
        input_ids = [self.tokenizer.token_to_id(x) for x in tokens]
        segment_ids = [0] * len(input_ids)
        attention_mask = [1] * len(input_ids)
        e = ExampleForTokenClassification(
            tokens=tokens,
            input_ids=input_ids,
            segment_ids=segment_ids,
            attention_mask=attention_mask,
            label_ids=[self.label_to_id(x) for x in instance["labels"]],
        )
        return e

    def parse_instances(self, instances, **kwargs) -> Iterator[ExampleForTokenClassification]:
        for instance in instances:
            e = self.parse_instance(instance, **kwargs)
            if not e:
                continue
            yield e



import re
from collections import namedtuple
from typing import Iterator

from tokenizers import BertWordPieceTokenizer

from naivenlp_datasets.readers import CsvFileReader, JsonlFileReader

ExampleForSequenceClassification = namedtuple(
    "ExampleForSequenceClassification",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "label"],
)


class JsonlFileReaderForSequenceClassification(JsonlFileReader):
    """Jsonl file reader for sequence classification"""

    def _parse_instance(
        self, data, sequence_column="sequence", label_column="label", with_pair=False, pair_column="pair", **kwargs
    ):
        instance = {
            "label": data[label_column],  # label is str, convert to in later
            "sequence": data[sequence_column],
        }
        if with_pair and pair_column in data:
            instance["pair"] = data[pair_column]
        return instance


class CsvFileReaderForSequenceClassification(CsvFileReader):
    """Csv file reader for sequene classification"""

    def _parse_instance(
        self, line, sep=",", sequence_column=0, label_column=-1, with_pair=False, pair_column=2, **kwargs
    ):
        parts = re.split(sep, line)
        instance = {
            "label": parts[label_column],
            "sequence": parts[sequence_column],
        }
        if with_pair:
            instance["pair"] = parts[pair_column]
        return instance


class ExampleParserForSequenceClassification:
    """Parse instance to ExampleForSequenceClassification"""

    def __init__(self, vocab_file, label_to_id=None, do_lower_case=True, **kwargs) -> None:
        self.tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)
        if label_to_id is None:
            label_to_id = lambda x: int(x)
        self.label_to_id = label_to_id

    def parse_instance(self, instance, add_special_tokens=True, **kwargs) -> ExampleForSequenceClassification:
        sequence, pair = instance["sequence"], instance.get("pair", None)
        encoding = self.tokenizer.encode(sequence, pair=pair, add_special_tokens=add_special_tokens)
        example = ExampleForSequenceClassification(
            tokens=encoding.tokens,
            input_ids=encoding.ids,
            segment_ids=encoding.type_ids,
            attention_mask=encoding.attention_mask,
            label=self.label_to_id(instance["label"]),
        )
        return example

    def parse_instances(
        self, instances, add_special_tokens=True, **kwargs
    ) -> Iterator[ExampleForSequenceClassification]:
        for instance in instances:
            example = self.parse_instance(instance, add_special_tokens=add_special_tokens, **kwargs)
            if not example:
                continue
            yield example



from typing import List

import torch
from naivenlp_datasets.sequence_classification import (
    CsvFileReaderForSequenceClassification,
    ExampleForSequenceClassification,
    ExampleParserForSequenceClassification,
    JsonlFileReaderForSequenceClassification,
)


def _padding_features(features, pad_id=0, max_length=512):
    padded_features = {}
    for feature in features:
        for k in feature.keys():
            if k not in padded_features:
                padded_features[k] = []
            value = feature[k]
            padded_value = value + [pad_id] * (max_length - len(value))
            padded_features[k].append(padded_value)
    # to tensor
    return {k: torch.LongTensor(v) for k, v in padded_features.items()}


class DatasetForSequenceClassification(torch.utils.data.Dataset):
    """Dataset for sequence classification in PyTorch"""

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, label_to_id=None, **kwargs):
        reader = JsonlFileReaderForSequenceClassification()
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, label_to_id=label_to_id, **kwargs)

    @classmethod
    def from_csv_files(cls, input_files, vocab_file, label_to_id=None, sep=",", **kwargs):
        reader = CsvFileReaderForSequenceClassification()
        instances = reader.read_files(input_files, sep=sep, **kwargs)
        return cls.from_instances(instances, vocab_file, label_to_id=label_to_id, **kwargs)

    @classmethod
    def from_tsv_files(cls, input_files, vocab_file, label_to_id=None, sep="\t", **kwargs):
        reader = CsvFileReaderForSequenceClassification()
        instances = reader.read_files(input_files, sep=sep, **kwargs)
        return cls.from_instances(instances, vocab_file, label_to_id=label_to_id, **kwargs)

    @classmethod
    def from_instances(
        cls,
        instances,
        vocab_file,
        label_to_id=None,
        max_sequence_length=512,
        do_lower_case=True,
        add_special_tokens=True,
        **kwargs
    ):
        parser = ExampleParserForSequenceClassification(
            vocab_file,
            label_to_id=label_to_id,
            max_sequence_length=max_sequence_length,
            do_lower_case=do_lower_case,
            **kwargs
        )
        examples = []
        for e in parser.parse_instances(instances, add_special_tokens=add_special_tokens, **kwargs):
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls, examples: List[ExampleForSequenceClassification], pad_id=0, max_sequence_length=512, **kwargs
    ):
        return cls(examples=examples, pad_id=pad_id, max_sequence_length=max_sequence_length, **kwargs)

    def __init__(
        self, examples: List[ExampleForSequenceClassification], pad_id=0, max_sequence_length=512, **kwargs
    ) -> None:
        super().__init__()
        self.examples = examples
        self.pad_id = pad_id
        self.max_sequence_length = max_sequence_length

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        e = self.examples[index]
        features = {
            "input_ids": e.input_ids,
            "segment_ids": e.segment_ids,
            "attention_mask": e.attention_mask,
        }
        labels = {
            "label": e.label,
        }
        return features, labels

    @property
    def batch_padding_collate(self):
        """Padding sequence to max length in a batch."""

        def _collate_fn(batch):
            max_length = max([len(x["input_ids"]) for x, _ in batch])
            features = _padding_features([x for x, _ in batch], pad_id=self.pad_id, max_length=max_length)
            labels = {"label": torch.LongTensor([y["label"] for _, y in batch])}
            return features, labels

        return _collate_fn

    @property
    def fixed_padding_collate(self):
        """Padding sequence to fixed length."""

        def _collate_fn(batch):
            features = _padding_features([x for x, _ in batch], pad_id=self.pad_id, max_length=self.max_sequence_length)
            labels = {"label": torch.LongTensor([y["label"] for _, y in batch])}
            return features, labels

        return _collate_fn



import torch


def padding_dict(datas, pad_id=0, max_length=512):
    padded_features = {}
    for data in datas:
        for k in data.keys():
            if k not in padded_features:
                padded_features[k] = []
            value = data[k]
            padded_value = value + [pad_id] * (max_length - len(value))
            padded_features[k].append(padded_value)
    # to tensor
    return {k: torch.LongTensor(v) for k, v in padded_features.items()}


def padding_features(features, pad_id=0, max_length=512):
    return padding_dict(features, pad_id=pad_id, max_length=max_length)


def padding_labels(labels, pad_id=0, max_length=512):
    return padding_dict(labels, pad_id=pad_id, max_length=max_length)



import logging
from typing import List

import tensorflow as tf
from naivenlp_datasets.simcse import (
    CsvFileReaderForSimCSE,
    ExampleForSimCSE,
    ExampleParserForSimCSE,
    JsonlFileReaderForSimCSE,
)

from . import utils
from .dataset import TFDataset


class TFDatasetForUnsupSimCSE(TFDataset):
    """Datapipe for unsupervised SimCSE"""

    def save_tfrecord(self, output_files, **kwargs):
        """Save examples to tf records"""

        def _encoding(example: ExampleForSimCSE):
            feature = {
                "input_ids": utils.int64_feature([int(x) for x in example.input_ids]),
                "segment_ids": utils.int64_feature([int(x) for x in example.segment_ids]),
                "attention_mask": utils.int64_feature([int(x) for x in example.attention_mask]),
            }
            return feature

        utils.save_tfrecord(self.examples, _encoding, output_files, **kwargs)

    @classmethod
    def from_tfrecord_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        dataset = utils.read_tfrecord_files(input_files, **kwargs)
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        # parse example
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, **kwargs) -> tf.data.Dataset:
        reader = JsonlFileReaderForSimCSE(with_pos_sequence=False, with_neg_sequence=False, **kwargs)
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_csv_files(cls, input_files, vocab_file, **kwargs):
        reader = CsvFileReaderForSimCSE(with_pos_sequence=False, with_neg_sequence=False, **kwargs)
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_instances(cls, instances, vocab_file, add_special_tokens=True, **kwargs) -> tf.data.Dataset:
        parser = ExampleParserForSimCSE(vocab_file, with_pos_sequence=False, with_neg_sequence=False, **kwargs)
        examples = []
        for e in parser.parse_instances(instances, add_special_tokens=add_special_tokens, **kwargs):
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls, examples: List[ExampleForSimCSE], return_self=False, verbose=True, **kwargs
    ) -> tf.data.Dataset:
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None
        if verbose:
            n = min(5, len(examples))
            for i in range(n):
                logging.info("Showing NO.%d example: %s", i, examples[i])

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
            )
        )
        # do transformation
        d = cls(examples=examples, **kwargs)
        if return_self:
            return d(dataset, **kwargs), d
        return d(dataset, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        if not do_filter:
            return dataset
        dataset = dataset.filter(lambda a, b, c: tf.size(a) <= max_sequence_length)
        return dataset

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        if not to_dict:
            dataset = dataset.map(
                lambda a, b, c: ((a, b, c), None),
                num_parallel_calls=num_parallel_calls,
            ).prefetch(buffer_size)
            return dataset
        dataset = dataset.map(
            lambda x, y, z: ({"input_ids": x, "segment_ids": y, "attention_mask": z}, None),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        return dataset

    def _fixed_padding(self, dataset: tf.data.Dataset, pad_id=0, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([maxlen, ], [maxlen, ], [maxlen, ])
        padding_values = (pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _batch_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _bucket_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id)
        # fmt: on
        return utils.bucketing_and_padding(
            dataset,
            bucket_fn=lambda a, b, c: tf.size(a),
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            **kwargs,
        )


class TFDatasetForSupervisedSimCSE(TFDataset):
    """Datapipe for unsupervised SimCSE"""

    def save_tfrecord(self, output_files, **kwargs):
        """Save examples in tfrecord format"""

        def _encoding(example: ExampleForSimCSE):
            feature = {
                "input_ids": utils.int64_feature([int(x) for x in example.input_ids]),
                "segment_ids": utils.int64_feature([int(x) for x in example.segment_ids]),
                "attention_mask": utils.int64_feature([int(x) for x in example.attention_mask]),
                "pos_input_ids": utils.int64_feature([int(x) for x in example.pos_input_ids]),
                "pos_segment_ids": utils.int64_feature([int(x) for x in example.pos_segment_ids]),
                "pos_attention_mask": utils.int64_feature([int(x) for x in example.pos_attention_mask]),
            }
            return feature

        utils.save_tfrecord(self.examples, _encoding, output_files, **kwargs)

    @classmethod
    def from_tfrecord_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        dataset = utils.read_tfrecord_files(input_files, **kwargs)
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        # parse example
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "pos_input_ids": tf.io.VarLenFeature(tf.int64),
            "pos_segment_ids": tf.io.VarLenFeature(tf.int64),
            "pos_attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_attention_mask"]), tf.int32),
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, **kwargs) -> tf.data.Dataset:
        reader = JsonlFileReaderForSimCSE(with_pos_sequence=True, with_neg_sequence=False, **kwargs)
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_csv_files(cls, input_files, vocab_file, **kwargs):
        reader = CsvFileReaderForSimCSE(with_pos_sequence=True, with_neg_sequence=False, **kwargs)
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_instances(cls, instances, vocab_file, add_special_tokens=True, **kwargs) -> tf.data.Dataset:
        parser = ExampleParserForSimCSE(vocab_file, with_pos_sequence=True, with_neg_sequence=False, **kwargs)
        examples = []
        for e in parser.parse_instances(instances, add_special_tokens=add_special_tokens, **kwargs):
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls, examples: List[ExampleForSimCSE], return_self=False, verbose=True, **kwargs
    ) -> tf.data.Dataset:
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None
        if verbose:
            n = min(5, len(examples))
            for i in range(n):
                logging.info("Showing NO.%d example: %s", i, examples[i])

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_attention_mask for e in examples], dtype=tf.int32),
            )
        )
        # do transformation
        d = cls(examples=examples, **kwargs)
        if return_self:
            return d(dataset, **kwargs), d
        return d(dataset, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        if not do_filter:
            return dataset
        dataset = dataset.filter(lambda a, b, c, d, e, f: tf.size(a) <= max_sequence_length)
        return dataset

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        if not to_dict:
            dataset = dataset.map(
                lambda a, b, c, d, e, f: ((a, b, c, d, e, f), None),
                num_parallel_calls=num_parallel_calls,
            ).prefetch(buffer_size)
            return dataset
        dataset = dataset.map(
            lambda a, b, c, d, e, f: (
                {
                    "input_ids": a,
                    "segment_ids": b,
                    "attention_mask": c,
                    "pos_input_ids": d,
                    "pos_segment_ids": e,
                    "pos_attention_mask": f,
                },
                None,
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        return dataset

    def _fixed_padding(self, dataset: tf.data.Dataset, pad_id=0, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _batch_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ], [None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _bucket_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ], [None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.bucketing_and_padding(
            dataset,
            bucket_fn=lambda a, b, c, d, e, f: tf.size(a),
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            **kwargs,
        )


class TFDatasetForHardNegativeSimCSE(TFDataset):
    """Datapipe for hard negative SimCSE"""

    def save_tfrecord(self, output_files, **kwargs):
        """Save examples in tfrecord format"""

        def _encoding(example: ExampleForSimCSE):
            feature = {
                "input_ids": utils.int64_feature([int(x) for x in example.input_ids]),
                "segment_ids": utils.int64_feature([int(x) for x in example.segment_ids]),
                "attention_mask": utils.int64_feature([int(x) for x in example.attention_mask]),
                "pos_input_ids": utils.int64_feature([int(x) for x in example.pos_input_ids]),
                "pos_segment_ids": utils.int64_feature([int(x) for x in example.pos_segment_ids]),
                "pos_attention_mask": utils.int64_feature([int(x) for x in example.pos_attention_mask]),
                "neg_input_ids": utils.int64_feature([int(x) for x in example.neg_input_ids]),
                "neg_segment_ids": utils.int64_feature([int(x) for x in example.neg_segment_ids]),
                "neg_attention_mask": utils.int64_feature([int(x) for x in example.neg_attention_mask]),
            }
            return feature

        utils.save_tfrecord(self.examples, _encoding, output_files, **kwargs)

    @classmethod
    def from_tfrecord_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        dataset = utils.read_tfrecord_files(input_files, **kwargs)
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        # parse example
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "pos_input_ids": tf.io.VarLenFeature(tf.int64),
            "pos_segment_ids": tf.io.VarLenFeature(tf.int64),
            "pos_attention_mask": tf.io.VarLenFeature(tf.int64),
            "neg_input_ids": tf.io.VarLenFeature(tf.int64),
            "neg_segment_ids": tf.io.VarLenFeature(tf.int64),
            "neg_attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["neg_input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["neg_segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["neg_attention_mask"]), tf.int32),
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, **kwargs) -> tf.data.Dataset:
        reader = JsonlFileReaderForSimCSE(with_pos_sequence=True, with_neg_sequence=True, **kwargs)
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_csv_files(cls, input_files, vocab_file, **kwargs):
        reader = CsvFileReaderForSimCSE(with_pos_sequence=True, with_neg_sequence=True, **kwargs)
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_instances(cls, instances, vocab_file, add_special_tokens=True, **kwargs) -> tf.data.Dataset:
        parser = ExampleParserForSimCSE(vocab_file, with_pos_sequence=True, with_neg_sequence=True, **kwargs)
        examples = []
        for e in parser.parse_instances(instances, add_special_tokens=add_special_tokens, **kwargs):
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(
        cls, examples: List[ExampleForSimCSE], return_self=False, verbose=True, **kwargs
    ) -> tf.data.Dataset:
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None
        if verbose:
            n = min(5, len(examples))
            for i in range(n):
                logging.info("Showing NO.%d example: %s", i, examples[i])

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.neg_input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.neg_segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.neg_attention_mask for e in examples], dtype=tf.int32),
            )
        )
        # do transformation
        d = cls(examples=examples, **kwargs)
        if return_self:
            return d(dataset, **kwargs), d
        return d(dataset, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        if not do_filter:
            return dataset
        dataset = dataset.filter(lambda a, b, c, d, e, f, g, h, i: tf.size(a) <= max_sequence_length)
        return dataset

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        if not to_dict:
            dataset = dataset.map(
                lambda a, b, c, d, e, f, g, h, i: ((a, b, c, d, e, f, g, h, i), None),
                num_parallel_calls=num_parallel_calls,
            ).prefetch(buffer_size)
            return dataset
        dataset = dataset.map(
            lambda a, b, c, d, e, f, g, h, i: (
                {
                    "input_ids": a,
                    "segment_ids": b,
                    "attention_mask": c,
                    "pos_input_ids": d,
                    "pos_segment_ids": e,
                    "pos_attention_mask": f,
                    "neg_input_ids": g,
                    "neg_segment_ids": h,
                    "neg_attention_mask": i,
                },
                None,
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        return dataset

    def _fixed_padding(self, dataset: tf.data.Dataset, pad_id=0, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _batch_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ], [None, ], [None, ], [None, ], [None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _bucket_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ], [None, ], [None, ], [None, ], [None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.bucketing_and_padding(
            dataset,
            bucket_fn=lambda a, b, c, d, e, f, g, h, i: tf.size(a),
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            **kwargs,
        )


class TFDatasetForSimCSE(TFDataset):
    """Dataset for SimCSE in TensorFlow"""

    def save_tfrecord(self, output_files, **kwargs):
        """save examples to tf records"""
        self.dataset.save_tfrecord(output_files, **kwargs)

    @classmethod
    def from_tfrecord_files(cls, input_files, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        if with_pos_sequence and with_neg_sequence:
            return TFDatasetForHardNegativeSimCSE.from_tfrecord_files(input_files, **kwargs)
        if with_pos_sequence:
            return TFDatasetForSupervisedSimCSE.from_tfrecord_files(input_files, **kwargs)
        return TFDatasetForUnsupSimCSE.from_tfrecord_files(input_files, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        if with_pos_sequence and with_neg_sequence:
            return TFDatasetForHardNegativeSimCSE.from_jsonl_files(input_files, vocab_file, **kwargs)
        if with_pos_sequence:
            return TFDatasetForSupervisedSimCSE.from_jsonl_files(input_files, vocab_file, **kwargs)
        return TFDatasetForUnsupSimCSE.from_jsonl_files(input_files, vocab_file, **kwargs)

    @classmethod
    def from_csv_files(
        cls, input_files, vocab_file, with_pos_sequence=False, with_neg_sequence=False, sep=",", **kwargs
    ):
        if with_pos_sequence and with_neg_sequence:
            return TFDatasetForHardNegativeSimCSE.from_csv_files(input_files, vocab_file, sep=sep, **kwargs)
        if with_pos_sequence:
            return TFDatasetForSupervisedSimCSE.from_csv_files(input_files, vocab_file, sep=sep, **kwargs)
        return TFDatasetForUnsupSimCSE.from_csv_files(input_files, vocab_file, sep=sep, **kwargs)

    @classmethod
    def from_instances(cls, instances, vocab_file, with_pos_sequence=False, with_neg_sequence=False, **kwargs):
        if with_pos_sequence and with_neg_sequence:
            return TFDatasetForHardNegativeSimCSE.from_instances(instances, vocab_file, **kwargs)
        if with_pos_sequence:
            return TFDatasetForSupervisedSimCSE.from_instances(instances, vocab_file, **kwargs)
        return TFDatasetForUnsupSimCSE.from_instances(instances, vocab_file, **kwargs)

    @classmethod
    def from_examples(
        cls,
        examples: List[ExampleForSimCSE],
        with_pos_sequence=False,
        with_neg_sequence=False,
        return_self=False,
        **kwargs
    ):
        if with_pos_sequence and with_neg_sequence:
            return TFDatasetForHardNegativeSimCSE.from_examples(examples, return_self=return_self, **kwargs)
        if with_pos_sequence:
            return TFDatasetForSupervisedSimCSE.from_examples(examples, return_self=return_self, **kwargs)
        return TFDatasetForUnsupSimCSE.from_examples(examples, return_self=return_self, **kwargs)

    def __init__(self, examples=None, with_pos_sequence=False, with_neg_sequence=False, **kwargs) -> None:
        super().__init__(examples=examples, **kwargs)
        self.with_pos_sequence = with_pos_sequence
        self.with_neg_sequence = with_neg_sequence
        if self.with_pos_sequence and self.with_neg_sequence:
            self.dataset = TFDatasetForHardNegativeSimCSE(examples=self.examples, **kwargs)
        elif self.with_pos_sequence:
            self.dataset = TFDatasetForSupervisedSimCSE(examples=self.examples, **kwargs)
        else:
            self.dataset = TFDatasetForUnsupSimCSE(examples=self.examples, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        return self.dataset._filter(dataset, do_filter=do_filter, max_sequence_length=max_sequence_length, **kwargs)

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        return self.dataset._to_dict(dataset, to_dict=to_dict, **kwargs)

    def _fixed_padding(self, dataset: tf.data.Dataset, pad_id=0, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        return self.dataset._fixed_padding(dataset, pad_id=pad_id, max_sequence_length=max_sequence_length, **kwargs)

    def _batch_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        return self.dataset._batch_padding(dataset, pad_id=pad_id, **kwargs)

    def _bucket_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        return self.dataset._bucket_padding(dataset, pad_id=pad_id, **kwargs)



from typing import List

from naivenlp_datasets.simcse import (
    CsvFileReaderForSimCSE,
    ExampleForSimCSE,
    ExampleParserForSimCSE,
    JsonlFileReaderForSimCSE,
)

from . import utils
from .dataset import PTDataset


class DatasetForSimCSE(PTDataset):
    """Dataset for SimCSE in PyTorch"""

    @classmethod
    def from_jsonl_files(
        cls,
        input_files,
        vocab_file,
        sequence_column="sequence",
        with_pos_sequence=False,
        with_neg_sequence=False,
        pos_sequence_column="pos_sequence",
        neg_sequence_column="neg_sequence",
        **kwargs
    ):
        reader = JsonlFileReaderForSimCSE(
            with_pos_sequence=with_pos_sequence,
            with_neg_sequence=with_neg_sequence,
            sequence_column=sequence_column,
            pos_sequence_column=pos_sequence_column,
            neg_sequence_column=neg_sequence_column,
            **kwargs
        )
        instances = reader.read_files(input_files, **kwargs)
        return cls.from_instances(
            instances, vocab_file, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )

    @classmethod
    def from_csv_files(
        cls,
        input_files,
        vocab_file,
        sep=",",
        sequence_column=0,
        with_pos_sequence=False,
        with_neg_sequence=False,
        pos_sequence_column=1,
        neg_sequence_column=2,
        **kwargs
    ):
        reader = CsvFileReaderForSimCSE(
            with_pos_sequence=with_pos_sequence,
            with_neg_sequence=with_neg_sequence,
            sequence_column=sequence_column,
            pos_sequence_column=pos_sequence_column,
            neg_sequence_column=neg_sequence_column,
            **kwargs
        )
        instances = reader.read_files(input_files, sep=sep, **kwargs)
        return cls.from_instances(
            instances, vocab_file, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )

    @classmethod
    def from_instances(
        cls, instances, vocab_file, with_pos_sequence=False, with_neg_sequence=False, add_special_tokens=True, **kwargs
    ):
        parser = ExampleParserForSimCSE(
            vocab_file, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )
        examples = []
        for e in parser.parse_instances(instances, add_special_tokens=add_special_tokens, **kwargs):
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(
            examples, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )

    @classmethod
    def from_examples(
        cls,
        examples: List[ExampleForSimCSE],
        with_pos_sequence=False,
        with_neg_sequence=False,
        max_sequence_length=512,
        **kwargs
    ):
        valid_examples = []
        for e in examples:
            if len(e.input_ids) > max_sequence_length:
                continue
            if with_pos_sequence and len(e.pos_input_ids) > max_sequence_length:
                continue
            if with_neg_sequence and len(e.neg_input_ids) > max_sequence_length:
                continue
            valid_examples.append(e)

        return cls(
            examples=valid_examples, with_pos_sequence=with_pos_sequence, with_neg_sequence=with_neg_sequence, **kwargs
        )

    def __init__(
        self, examples: List[ExampleForSimCSE], with_pos_sequence=False, with_neg_sequence=False, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.examples = examples
        self.with_pos_sequence = with_pos_sequence
        self.with_neg_sequence = with_neg_sequence

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        e = self.examples[index]
        features = {
            "input_ids": e.input_ids,
            "segment_ids": e.segment_ids,
            "attention_mask": e.attention_mask,
        }
        if self.with_pos_sequence:
            features.update(
                {
                    "pos_input_ids": e.pos_input_ids,
                    "pos_segment_ids": e.pos_segment_ids,
                    "pos_attention_mask": e.pos_attention_mask,
                }
            )
        if self.with_neg_sequence:
            features.update(
                {
                    "neg_input_ids": e.neg_input_ids,
                    "neg_segment_ids": e.neg_segment_ids,
                    "neg_attention_mask": e.neg_attention_mask,
                }
            )
        return features, None

    @property
    def batch_padding_collate(self):
        """Padding sequence to max length in batch"""

        def _collate_fn(batch):
            max_length = 0
            max_length = max([max_length] + [len(x["input_ids"]) for x, _ in batch])
            if self.with_pos_sequence:
                max_length = max([max_length] + [len(x["pos_input_ids"]) for x, _ in batch])
            if self.with_neg_sequence:
                max_length = max([max_length] + [len(x["neg_input_ids"]) for x, _ in batch])
            padded_features = utils.padding_features([x for x, _ in batch], pad_id=self.pad_id, max_length=max_length)
            return padded_features, None

        return _collate_fn

    @property
    def fixed_padding_collate(self):
        """Padding sequence to a fixed length"""

        def _collate_fn(batch):
            max_length = self.max_sequence_length
            padded_features = utils.padding_features([x for x, _ in batch], pad_id=self.pad_id, max_length=max_length)
            return padded_features, None

        return _collate_fn



import unittest

LABEL_MAP = {
    "O": 0,
    "B-BRAND": 1,
    "I-BRAND": 2,
}


def _label_to_id(label):
    return LABEL_MAP[label]


class DatasetTest(unittest.TestCase):
    """Dataset test for token classification"""

    def test_dataset_tf(self):
        from naivenlp_datasets.tf import TFDatasetForTokenClassification

        dataset, d = TFDatasetForTokenClassification.from_jsonl_files(
            input_files=["testdata/token_classification.jsonl"],
            vocab_file="testdata/vocab.txt",
            label_to_id=_label_to_id,
            return_self=True,
            batch_size=4,
        )
        for idx, batch in enumerate(iter(dataset)):
            print()
            print("No.{} batch:\n{}".format(idx, batch))

        d.save_tfrecord("testdata/token_classification.tfrecord")

        dataset = TFDatasetForTokenClassification.from_tfrecord_files(
            input_files="testdata/token_classification.tfrecord",
            batch_size=4,
        )

    def test_dataset_pt(self):
        import torch
        from naivenlp_datasets.pt import DatasetForTokenClassification

        dataset = DatasetForTokenClassification.from_jsonl_files(
            input_files=["testdata/token_classification.jsonl"],
            vocab_file="testdata/vocab.txt",
            label_to_id=_label_to_id,
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, num_workers=1, batch_size=4, collate_fn=dataset.batch_padding_collate
        )
        for idx, batch in enumerate(dataloader):
            print()
            print("No.{} batch:\n{}".format(idx, batch))


if __name__ == "__main__":
    unittest.main()
